window.Gleam.initSearch([{"doc":"chip","title":"chip","content":"# chip\n\n[![Package Version](https://img.shields.io/hexpm/v/chip)](https://hex.pm/packages/chip)\n[![Hex Docs](https://img.shields.io/badge/hex-docs-ffaff3)](https://hexdocs.pm/chip/)\n\n\nA local process registry that plays along with Gleam&#39;s [Subject](https://hexdocs.pm/gleam_erlang/gleam/erlang/process.html#Subject) type. \n\nIt can hold a set of subjects to later retrieve individually or dispatch as a group. Will also automatically delist dead processes.\n\n### Example\n\nLets assemble the pieces to build a simple counter actor:\n\n```gleam\nimport gleam/erlang/process\nimport gleam/otp/actor\n\npub opaque type Message {\n  Inc\n  Current(client: process.Subject(Int))\n}\n\npub fn increment(counter) {\n  actor.send(counter, Inc)\n}\n\npub fn current(counter) {\n  actor.call(counter, Current(_), 10)\n}\n\nfn loop(message: Message, count: Int) {\n  case message {\n    Inc -&gt; {\n      actor.Continue(count + 1, option.None)\n    }\n\n    Current(client) -&gt; {\n      process.send(client, count)\n      actor.Continue(count, option.None)\n    }\n  }\n}\n```\n\nWe start our registry and create new instances of a counter:\n\n```gleam\nimport gleam/otp/actor\nimport chip/group\n\npub fn main() {\n  let assert Ok(registry) = chip.start()\n\n  let assert Ok(counter_1) = actor.start(0, loop)\n  let assert Ok(counter_2) = actor.start(0, loop)\n  let assert Ok(counter_3) = actor.start(0, loop)\n\n  chip.register(registry, chip.new(counter_1) |&gt; chip.tag(1))\n  chip.register(registry, chip.new(counter_2) |&gt; chip.tag(2))\n  chip.register(registry, chip.new(counter_3) |&gt; chip.tag(3))\n  \n  process.sleep_forever()\n}\n```\n\nLater, we may retrieve a member:\n \n```gleam\nlet assert Ok(counter) = chip.find(registry, 2)\nlet assert 0 = counter.current(counter)\n```\n\nOr broadcast a message to all members:\n\n```gleam\nchip.dispatch(registry, fn(counter) {\n  actor.increment(counter)\n}) \n\nlet assert Ok(counter) = chip.find(registry, 2)\nlet assert 1 = counter.current(counter)\n```\n\n## The road towards V1\n\nFeature-wise this is near beign complete. Still planning to integrate: \n\n- [x] Adjust the API to be more in-line with other Registry libraries like elixir&#39;s Registry, erlang&#39;s pg or Syn. \n- [x] Document modules.\n- [x] Basic test case scenarios.\n- [X] Should play well with gleam style of supervisors. \n- [ ] Document guides and use-cases. \n- [ ] Build a benchmark that measures performance and memory consuption. \n- [ ] Implement an ETS backend. \n- [ ] Benchmark in-process backend vs ETS backend.\n\n## Previous Art\n\nThis registry takes and combines some ideas from Elixir’s [Registry](https://hexdocs.pm/elixir/Kernel.html), Erlang’s [pg](https://www.erlang.org/doc/apps/kernel/pg.html) and [Syn](https://github.com/ostinelli/syn).\n\n## Alternatives\n\n[Singularity](https://hexdocs.pm/singularity/) is a gleam library that offers registry capabilities but focusing more on singleton actors, therefore it is better suited for keeping track of actors that need to be passed around as configuration through your app. \n\n## Installation\n\n```sh\ngleam add chip\n```\n","url":"index.html"},{"doc":"chip","title":"chip","content":"# General use cases \n\nChip is meant as a general solution for naming and grouping processess, therefore there may be many use cases attached to it. \n\n## As a process Index\n\nProbably one of the most common uses cases for chip will involve looking up for subjects in your system. \n\nFor example, in a web app we may add the registry as part of our app context. \n\n```gleam\npub fn main() {\n  let assert Ok(registry) = chip.start()\n  let self: process.Subject(Nil) = process.new_subject()\n  \n  chip.register(registry, chip.new(self) |&gt; chip.tag(&quot;1st&quot;))\n  chip.register(registry, chip.new(self) |&gt; chip.tag(&quot;2nd&quot;))\n  chip.register(registry, chip.new(self) |&gt; chip.tag(&quot;3rd&quot;))\n  \n  web.server(Context(registry))\n  \n  process.sleep_forever()\n}\n```\n\nThen have access to the subjects at a completely different scope.\n\n```gleam\npub server(request, context) {\n  case request.route {\n    Get, [&quot;/resource/&quot;, id] -&gt; {\n      let result = chip.find(context.registry, id)\n      render(result)\n    }\n  }\n}\n```\n\nOf course, this ability to directly reference subjects is not very useful without a supervision tree, as if the subject dies we can no longer send messages to it. Check the [supervision guideline](chip-as-part-of-a-supervision-tree.html) for more. \n\n## As a dispatcher\n\nOne of the most useful use cases for chip involves dispatching messages to a collection of subjects. \n\nFor example, we may register subjects as part of a group: \n\n```gleam\ntype Group {\n  Even\n  Odd\n}\n\npub fn main() {\n  let assert Ok(registry) = chip.start()\n\n  let assert Ok(store_1) = store.start(1)\n  let assert Ok(store_2) = store.start(2)\n  let assert Ok(store_3) = store.start(3)\n\n  chip.register(registry, chip.new(store_1) |&gt; chip.group(Odd))\n  chip.register(registry, chip.new(store_2) |&gt; chip.group(Even))\n  chip.register(registry, chip.new(store_3) |&gt; chip.group(Odd))\n}\n``` \n\nThen later dispatch messages to all members:\n\n```gleam\nchip.dispatch(registry, fn(store) {\n  store.increment(store)   \n})\n```\n\nOr a sub-group of members:\n\n```gleam\nchip.dispatch_group(registry, Odd, fn(store) {\n  store.increment(store)   \n})\n```\n\nThis is specially useful when wanting to create a PubSub system between subjects. Check the [PubSub Guideline](chip-as-a-local-pubsub.html) for more.  \n\n## As app configuration\n\nOther libraries in the erlang ecosystem like [registry](https://hexdocs.pm/elixir/Kernel.html), [pg](https://www.erlang.org/doc/apps/kernel/pg.html) and [syn](https://github.com/ostinelli/syn) can serve the purpose of storing configuration along processess.\n\nChip may not be well suited for this purpose as it can only store subjects of a single message type. If you need to reference subjects with different message types you may look at the [singularity](https://hexdocs.pm/singularity/) library. Also check the the [wrapping up chip Guideline](wrapping-up-chip.html) for how you may re-purpose a registry for this.\n","url":"guides/general-use-cases.html"},{"doc":"chip","title":"chip","content":"# Chip as part of a Supervision tree\n\nA supervision tree is a strategy used in the wider erlang ecosystem to keep long-running processeess alive. When a process in any part of the tree terminates the supervisor will attempt to restart the process and any subsequent processess after it. Giving our whole system self-healing properties.\n\nLets assume we have the simplest &quot;counter&quot; actor code: \n\n```gleam\npub opaque type Message {\n  Inc\n  Oops\n}\n\nfn loop(message: Message, count: Int) {\n  case message {\n    Inc -&gt; actor.continue(count + 1)\n    Oops -&gt; panic as &quot;unexpected error&quot;\n  }\n}\n```\n\nIntegrating to a supervisor is simple enough if we create a `childspec` for our counter: \n\n```gleam \npub fn main() {\n  let self = process.new_subject()\n\n  let childspec = fn(_param) {\n    // start the counter\n    use counter &lt;- try(actor.start(0, loop))\n    // on success, send counter back to caller.\n    process.send(self, counter)\n    Ok(counter)\n  }\n\n  // start all processess under a supervision tree \n  let assert Ok(_supervisor) =\n    supervisor.start(fn(children) {\n      children\n      |&gt; supervisor.add(supervisor.worker(childspec))\n    })\n\n  do_work(self)\n}\n```\n\nWith the implementation above, we have asured that our supervisor will keep this specific counter spec alive. In the case of unexpected termination the supervisor will re-run the childspec again to restore this and follow-up processess.  \n\nFor example lets assume the counter is terminated at the `do_work` function: \n\n```gleam\nfn do_work(self) {\n  // wait to receive the counter&#39;s subject, and operate on it\n  let assert Ok(counter) = process.receive(self, 50)\n  io.debug(counter)\n\n  // lets attempt to restart it and receive another reference\n  process.send(counter, Oops)\n  let assert Ok(counter) = process.receive(self, 50)\n  io.debug(counter)\n}\n```\n\nThe two subject counter references printed to the terminal are: \n\n```erlang\nSubject(//erl(&lt;0.90.0&gt;), //erl(#Ref&lt;0.3359404744.2770337801.190515&gt;))\nSubject(//erl(&lt;0.91.0&gt;), //erl(#Ref&lt;0.3359404744.2770337801.190559&gt;))\n```\n\nWe can confirm that both process ids `0.90.0` and `0.91.0` are different, therefore when the first counter was terminated the supervisor restarted it and created `0.91.0` in its place.\n\nAll of this requires carrying the `self` reference around the program and knowing when to receive the process in case of failure. We may be completely out of scope. This is where a registry may help. \n\nLets build a `start` specification for our counter actor: \n\n```gleam\npub fn start(registry, tag) {\n  let init = fn() { init(registry, tag) }\n  actor.start_spec(actor.Spec(init: init, init_timeout: 10, loop: loop))\n}\n\nfn init(registry, id) {\n  // Create a reference to self\n  let self = process.new_subject()\n\n  // Register the counter under an id on initialization\n  chip.register(\n    registry,\n    self\n      |&gt; chip.new()\n      |&gt; chip.tag(id),\n  )\n\n  // Adding self to the selector allows us to receive the Stop message\n  actor.Ready(\n    0,\n    process.new_selector()\n      |&gt; process.selecting(self, function.identity),\n  )\n}\n```\n\nThen integrate chip and our new start function under our supervisor: \n\n```gleam\npub fn main() {\n  let self = process.new_subject()\n\n  let childspec_registry = fn(_param) {\n    use registry &lt;- try(chip.start())\n    // on success, send the registry back to caller.\n    process.send(self, registry)\n    Ok(registry)\n  }\n\n  // Transform initial child parameter to the registry and an id tag\n  let updater_registry = fn(_param, registry) { #(registry, 1) }\n\n  let childspec_counter = fn(param) {\n    // We now receive the registry and initial id\n    let #(registry, id) = param\n    start(registry, id)\n  }\n\n  // Subsequent child counters will increment their id tag\n  let updater_counter = fn(param, _counter) {\n    let #(registry, id) = param\n    #(registry, id + 1)\n  }\n\n  // start all processess under a supervision tree \n  let assert Ok(_supervisor) =\n    supervisor.start(fn(children) {\n      children\n      |&gt; supervisor.add(\n        supervisor.worker(childspec_registry)\n        |&gt; supervisor.returning(updater_registry),\n      )\n      |&gt; supervisor.add(\n        supervisor.worker(childspec_counter)\n        |&gt; supervisor.returning(updater_counter),\n      )\n    })\n\n  // wait to receive the registry&#39;s subject, and operate on it\n  let assert Ok(registry) = process.receive(self, 50)\n  do_work(registry)\n}\n```\n\nIt is quite a bit of extra code and specification, but keep in mind you may move most of this to helpers. Now we can reference the counter from the registry in our `do_work` function.\n\n```gleam\nfn do_work(registry) {\n  // retrieve the counter&#39;s subject, and operate on it\n  let assert Ok(counter) = chip.find(registry, 1)\n  io.debug(counter)\n\n  // lets attempt to restart it and wait for the registry to update\n  process.send(counter, Oops)\n  process.sleep(50)\n  let assert Ok(counter) = chip.find(registry, 1)\n  io.debug(counter)\n}\n```\n\nPrinting us these subjects: \n\n```erlang\nSubject(//erl(&lt;0.91.0&gt;), //erl(#Ref&lt;0.211861637.3582984195.76640&gt;))\nSubject(//erl(&lt;0.92.0&gt;), //erl(#Ref&lt;0.211861637.3582984195.76689&gt;))\n```\n\nGranted the registry didn&#39;t solve all our issues, now we need to pass around a copy of our registry&#39;s subject through our app. And the registry may not yet have registered the new subject (we had to wait a few milliseconds for it to update and restart). \n\nThese issues are out of scope of chip but may be solved through different techniques. For example, top-level processes may use an &quot;app configuration&quot; library that keeps track of singleton processes, if you&#39;d like to re-purpose chip for this please check the [wrapping up chip Guideline](wrapping-up-chip.html) for more.\n","url":"guides/chip-as-part-of-a-supervision-tree.html"},{"doc":"chip","title":"chip","content":"# Chip as a local PubSub system","url":"guides/chip-as-a-local-pubsub.html"},{"doc":"chip","title":"chip","content":"# Wrapping up chip\n\n","url":"guides/wrapping-up-chip.html"},{"doc":"chip","title":"Chip","content":"pub opaque type Chip(msg, tag, group)\n A &quot;chip&quot; used for registration. Check the [new](#new) function.\n\nSynonyms:\nchip.Chip\nchip Chip","url":"chip.html#Chip"},{"doc":"chip","title":"Message","content":"pub opaque type Message(msg, tag, group)\n Chip&#39;s internal message type.\n\nSynonyms:\nchip.Message\nchip Message","url":"chip.html#Message"},{"doc":"chip","title":"Registry","content":"pub type Registry(msg, tag, group) =\n  process.Subject(Message(msg, tag, group))\n An shorter alias for the registry&#39;s Subject. \n \n Sometimes, when building out your system it may be useful to state the Registry&#39;s types.\n \n ## Example\n \n \n let assert Ok(registry) = chip.start()\n let registry: chip.Registry(Event, Id, Topic)\n \n \n Which is equivalent to: \n \n \n let assert Ok(registry) = chip.start()\n let registry: process.Subject(chip.Message(Event, Id, Topic))\n \n \n By specifying the types we can document the kind of registry we are working with.\n\nSynonyms:\nchip.Registry\nchip Registry","url":"chip.html#Registry"},{"doc":"chip","title":"dispatch","content":"pub fn dispatch(\n  registry: Subject(Message(a, b, c)),\n  callback: fn(Subject(a)) -&gt; d,\n) -&gt; Nil\n Applies a callback over all registered Subjects.\n \n ## Example\n \n \n chip.dispatch(registry, fn(subject) { \n   process.send(subject, message)\n })\n \nSynonyms:\nchip.dispatch\nchip dispatch","url":"chip.html#dispatch"},{"doc":"chip","title":"dispatch_group","content":"pub fn dispatch_group(\n  registry: Subject(Message(a, b, c)),\n  group: c,\n  callback: fn(Subject(a)) -&gt; d,\n) -&gt; Nil\n Applies a callback over a group.\n \n ## Example\n \n \n chip.dispatch_group(registry, Pets, fn(subject) { \n   process.send(subject, message)\n })\n \nSynonyms:\nchip.dispatch_group\nchip dispatch_group","url":"chip.html#dispatch_group"},{"doc":"chip","title":"find","content":"pub fn find(\n  registry: Subject(Message(a, b, c)),\n  tag: b,\n) -&gt; Result(Subject(a), Nil)\n Retrieves a tagged subject.\n \n ## Example\n \n \n let assert Ok(subject) = chip.find(registry, &quot;Luis&quot;)\n \nSynonyms:\nchip.find\nchip find","url":"chip.html#find"},{"doc":"chip","title":"group","content":"pub fn group(\n  registrant: Chip(a, b, c),\n  group: c,\n) -&gt; Chip(a, b, c)\n Adds the registrant under a group. \n \n ## Example\n \n \n chip.new(subject)\n |&gt; chip.group(General) \n \nSynonyms:\nchip.group\nchip group","url":"chip.html#group"},{"doc":"chip","title":"new","content":"pub fn new(subject: Subject(a)) -&gt; Chip(a, b, c)\n Creates a new registrant value. \n \n ## Example\n \n \n chip.new(subject)\n \nSynonyms:\nchip.new\nchip new","url":"chip.html#new"},{"doc":"chip","title":"register","content":"pub fn register(\n  registry: Subject(Message(a, b, c)),\n  registrant: Chip(a, b, c),\n) -&gt; Nil\n Registers a `Registrant`. \n \n ## Example\n \n \n let assert Ok(registry) = chip.start()\n \n chip.new(subject)\n |&gt; chip.register(registry, _)\n \n \n `Registrant` may be registered under a tag or group.\n \n \n let assert Ok(registry) = chip.start()\n \n chip.new(subject)\n |&gt; chip.tag(&quot;Francisco&quot;) \n |&gt; chip.group(Coffee) \n |&gt; chip.register(registry, _)\n \n \n You may register any subject at any point in time but usually keeping it under the initialization\n step of your process (like an Actor&#39;s `init` callback) will keep things organized and tidy. \nSynonyms:\nchip.register\nchip register","url":"chip.html#register"},{"doc":"chip","title":"start","content":"pub fn start() -&gt; Result(Subject(Message(a, b, c)), StartError)\n Starts the registry.\n \n ## Example\n \n \n &gt; chip.start()\n \nSynonyms:\nchip.start\nchip start","url":"chip.html#start"},{"doc":"chip","title":"tag","content":"pub fn tag(registrant: Chip(a, b, c), tag: b) -&gt; Chip(a, b, c)\n Adds a unique tag to a registrant, it will overwrite any previous subject under the same tag.\n \n ## Example\n \n \n chip.new(subject)\n |&gt; chip.tag(&quot;Luis&quot;) \n \nSynonyms:\nchip.tag\nchip tag","url":"chip.html#tag"},{"doc":"chip","title":"chip","content":" Chip is a local process registry that plays along with Gleam&#39;s `Subject` type for referencing\n erlang processes. It can hold to a set of subjects to later reference individually or dispatch \n a callback as a group. Will also automatically delist dead processes.","url":"chip.html"}]);